// Copyright (c) 1999-2012  Pavel Rychly

#ifndef FINLIB_LEXICON_HH
#define FINLIB_LEXICON_HH

#include <finlib/generator.hh>
#include <finlib/binfile.hh>

typedef unsigned int lexpos;

class lexicon {
protected:
    lexicon () {}
public:
    virtual ~lexicon();
    virtual int size() = 0;
    virtual const char* id2str (int id) = 0;
    virtual int str2id (const char *str) = 0;
    virtual Generator<int> *pref2ids (const char *str);
    virtual Generator<int> *regex2ids (const char *pattern);
};


class plain_lexicon: public lexicon  {
public:
    plain_lexicon (const char *filename);
    virtual int size();
    virtual const char *id2str (int id);
    virtual int str2id (const char *str);
    virtual Generator<int> *pref2ids (const char *str);
    virtual ~plain_lexicon();
    int srtidx2id (int sort_index_position);
    bool havetoswap;
protected:
    FILE *lexf;
    BinFile<int> *lidxf;
    BinFile<int> *lsrtf;
};


template <class intStore>
class gen_map_lexicon
{
protected:
    typedef typename intStore::const_iterator Iterator;
public:
    class IdsIterator : public Generator<int> {
        Iterator it;
        int ids_count;
    public:
        IdsIterator (Iterator i, int c): it(i), ids_count (c) {}
        virtual int next() {
            if (ids_count) {
                ids_count--; return *it++;
            } else
                return *it;
        }
        virtual bool end() {return ids_count == 0;}
        virtual int size() {return ids_count;}
    };
    gen_map_lexicon (const std::string &filename);
    int size()  {return lidxf.size();}
    const char *id2str (int id) {return id < 0 ? "" : lexf.at (lidxf [id]);}
    int str2id (const char *str);
    Generator<int> *pref2ids (const char *str);
    int srtidx2id (int sort_index_pos) {return lsrtf [sort_index_pos];}
protected:
    MapBinFile<char> lexf;
    MapBinFile<lexpos> lidxf;
    intStore lsrtf;
};

typedef gen_map_lexicon<MapBinFile<int> > map_lexicon;
typedef gen_map_lexicon<MapNetIntFile> net_map_lexicon;

template <class intStore>
gen_map_lexicon<intStore>::gen_map_lexicon (const std::string &filename)
    : lexf (filename + ".lex"), lidxf (filename + ".lex.idx"),
      lsrtf (filename + ".lex.srt")
{
}

template <class intStore>
int gen_map_lexicon<intStore>::str2id (const char *str)
{
    int minid = -1;
    int maxid = size();
    int m, c, id;
    while (minid < maxid -1) {
        m = (minid + maxid) /2;
        c = strcmp (id2str (id = lsrtf [m]), str);
        if (c == 0) {
            return id;
        } else if (c < 0) {
            minid = m;
        } else {
            maxid = m;
        }
    }
    return -1;
}

template <class intStore>
Generator<int>*
gen_map_lexicon<intStore>::pref2ids (const char *str)
{
    //if (*str == '\0') return lexicon::pref2ids (str);
    int minid = -1;
    int maxid = size();
    int m, c, beg, end;
    int len = strlen (str);
    // pokud je size() == -1 (asi by nemelo nastat), musime 
    //    inicializovat beg, end
    beg = 0;
    end = -1;
    while (minid < maxid -1) {
        m = (minid + maxid) /2;
        c = strncmp (str, id2str (lsrtf [m]), len);
        if (c == 0) {
            beg = end = m;
            break;
        } else if (c > 0) {
            minid = m;
        } else {
            maxid = m;
        }
    }
    if (minid >= maxid -1)
        return new IdsIterator (lsrtf.at (0), 0);

    /* najdeme spodni hranici */
    while (minid < beg -1) {
        m = (minid + beg) /2;
        if (strncmp (str, id2str (lsrtf [m]), len) == 0) {
            beg = m;
        } else {
            minid = m;
        }
    }
    /* najdeme horni hranici */
    while (end < maxid -1) {
        m = (end + maxid) /2;
        if (strncmp (str, id2str (lsrtf [m]), len) == 0) {
            end = m;
        } else {
            maxid = m;
        }
    }
    return new IdsIterator (lsrtf.at (beg), end - beg +1);
}


#endif

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:
